@@ -167,28 +167,89 @@ class XMLRPCHandlerTests(unittest.TestCase):
         self.assertEqual(request._error, 500)
 
 class TraverseTests(unittest.TestCase):
-    def test_underscore(self):
+    def test_security_disallows_underscore_methods(self):
         from supervisor import xmlrpc
-        self.assertRaises(xmlrpc.RPCError, xmlrpc.traverse, None, '_', None)
+        class Root:
+            pass
+        class A:
+            def _danger(self):
+                return True
+        root = Root()
+        root.a = A()
+        self.assertRaises(xmlrpc.RPCError, xmlrpc.traverse,
+            root, 'a._danger', [])
 
-    def test_notfound(self):
+    def test_security_disallows_object_traversal(self):
         from supervisor import xmlrpc
-        self.assertRaises(xmlrpc.RPCError, xmlrpc.traverse, None, 'foo', None)
+        class Root:
+            pass
+        class A:
+            pass
+        class B:
+            def danger(self):
+                return True
+        root = Root()
+        root.a = A()
+        root.a.b = B()
+        self.assertRaises(xmlrpc.RPCError, xmlrpc.traverse,
+            root, 'a.b.danger', [])
 
-    def test_badparams(self):
+    def test_namespace_name_not_found(self):
         from supervisor import xmlrpc
-        self.assertRaises(xmlrpc.RPCError, xmlrpc.traverse, self,
-                          'test_badparams', (1, 2, 3))
+        class Root:
+            pass
+        root = Root()
+        self.assertRaises(xmlrpc.RPCError, xmlrpc.traverse,
+            root, 'notfound.hello', None)
+
+    def test_method_name_not_found(self):
+        from supervisor import xmlrpc
+        class Root:
+            pass
+        class A:
+            pass
+        root = Root()
+        root.a = A()
+        self.assertRaises(xmlrpc.RPCError, xmlrpc.traverse,
+            root, 'a.notfound', [])
+
+    def test_method_name_exists_but_is_not_a_method(self):
+        from supervisor import xmlrpc
+        class Root:
+            pass
+        class A:
+            pass
+        class B:
+            pass
+        root = Root()
+        root.a = A()
+        root.a.b = B()
+        self.assertRaises(xmlrpc.RPCError, xmlrpc.traverse,
+            root, 'a.b', [])  # b is not a method
+
+    def test_bad_params(self):
+        from supervisor import xmlrpc
+        class Root:
+            pass
+        class A:
+            def hello(self, name):
+                return "Hello %s" % name
+        root = Root()
+        root.a = A()
+        self.assertRaises(xmlrpc.RPCError, xmlrpc.traverse,
+            root, 'a.hello', ["there", "extra"])  # too many params
 
     def test_success(self):
         from supervisor import xmlrpc
-        L = []
-        class Dummy:
-            def foo(self, a):
-                L.append(a)
-        dummy = Dummy()
-        xmlrpc.traverse(dummy, 'foo', [1])
-        self.assertEqual(L, [1])
+        class Root:
+            pass
+        class A:
+            def hello(self, name):
+                return "Hello %s" % name
+        root = Root()
+        root.a = A()
+        result = xmlrpc.traverse(root, 'a.hello', ["there"])
+        self.assertEqual(result, "Hello there")
 
 class SupervisorTransportTests(unittest.TestCase):
     def _getTargetClass(self):
@@ -297,42 +358,57 @@ class IterparseLoadsTests(unittest.TestCase):
         <value><i4>41</i4></value>
         </param>
         <param>
-        <value><int>14</int></value>
+        <value><string>foo</string></value>
+        </param>
+        <param>
+        <value><string></string></value>
+        </param>
+        <param>
+        <!-- xml-rpc spec allows strings without <string> tag -->
+        <value>bar</value>
+        </param>
+        <param>
+        <value></value>
         </param>
         <param>
         <value><boolean>1</boolean></value>
         </param>
         <param>
-        <value><string>hello world</string></value>
-        </param>
-        <param>
         <value><double>-12.214</double></value>
         </param>
         <param>
-        <value><dateTime.iso8601>19980717T14:08:55</dateTime.iso8601></value>
+        <value>
+        <dateTime.iso8601>19980717T14:08:55</dateTime.iso8601>
+        </value>
         </param>
         <param>
         <value><base64>eW91IGNhbid0IHJlYWQgdGhpcyE=</base64></value>
         </param>
         <param>
         <struct>
-          <member><name>k</name><value><i4>5</i4></value></member>
+        <member><name>j</name><value><i4>5</i4></value></member>
+        <member><name>k</name><value>abc</value></member>
         </struct>
         </param>
         <param>
         <array>
           <data>
             <value><i4>12</i4></value>
+            <value><string>abc</string></value>
+            <value>def</value>
             <value><i4>34</i4></value>
           </data>
         </array>
         </param>
         <param>
         <struct>
-        <member>
-          <name>k</name>
-          <value><array><data><value><i4>1</i4></value></data></array></value>
-        </member>
+          <member>
+            <name>k</name>
+            <value><array><data>
+              <value><i4>1</i4></value>
+              <struct></struct>
+            </data></array></value>
+          </member>
         </struct>
         </param>
         </params>
@@ -346,15 +422,17 @@ class IterparseLoadsTests(unittest.TestCase):
         import datetime
         self.assertEqual(method, 'examples.getStateName')
         self.assertEqual(params[0], 41)
-        self.assertEqual(params[1], 14)
-        self.assertEqual(params[2], True)
-        self.assertEqual(params[3], 'hello world')
-        self.assertEqual(params[4], -12.214)
-        self.assertEqual(params[5], datetime.datetime(1998, 7, 17, 14, 8, 55))
-        self.assertEqual(params[6], "you can't read this!")
-        self.assertEqual(params[7], {'k': 5})
-        self.assertEqual(params[8], [12, 34])
-        self.assertEqual(params[9], {'k': [1]})
+        self.assertEqual(params[1], 'foo')
+        self.assertEqual(params[2], '')
+        self.assertEqual(params[3], 'bar')
+        self.assertEqual(params[4], '')
+        self.assertEqual(params[5], True)
+        self.assertEqual(params[6], -12.214)
+        self.assertEqual(params[7], datetime.datetime(1998, 7, 17, 14, 8, 55))
+        self.assertEqual(params[8], "you can't read this!")
+        self.assertEqual(params[9], {'j': 5, 'k': 'abc'})
+        self.assertEqual(params[10], [12, 'abc', 'def', 34])
+        self.assertEqual(params[11], {'k': [1, {}]})
 
 class DummyResponse:
     def __init__(self, status=200, body='', reason='reason'):
